""" global constants """
# for shortest path calculation
MAX_LABEL_COST = 2147483647
# in case where the divisor/demoninator is ZERO
SMALL_DIVISOR = 0.001
# for column generation
MIN_OD_VOL = 0.000001
# for accessibility evaluation
MIN_TIME_BUDGET = 10
MAX_TIME_BUDGET = 240
BUDGET_TIME_INTVL = 5
# reserved for simulation
NUM_OF_SECS_PER_SIMU_INTERVAL = 6